@@ -11,8 +11,8 @@ automake=automake
 autoheader=autoheader
 aclocal=aclocal
 
-libtoolize --copy --force
+libtoolize --copy --force --ltdl
 $aclocal
 $autoheader
-$automake --verbose --gnu --add-missing --copy --force-missing
+$automake --gnu --add-missing --copy --force-missing
 $autoconf
