@@ -31,6 +31,7 @@
 #ifndef _ISO15693_H_
 #define _ISO15693_H_
 
+#include <db.h>
 #include <rfid_reader.h>
 
 /*
@@ -48,6 +49,10 @@
  * not included in the distribution of this software. To comply with
  * the licensing terms of the document, I'll destroy all digital
  * copies of the document I have after giving it to someone.
+ * 
+ * Alternatively, anyone can get the final draft of the standard at
+ * http://www.wg8.de/sd1.html. It is not the same document but it does
+ * not differ much.
  *
  * ISO-15693 is NOT a standard covering host <-> reader protocol but
  * reader <-> transponder protocol. However it is sometime used as a
@@ -75,19 +80,18 @@
  */
 
 typedef struct iso15693 {
-  u_int8_t flags;		/* Permanent flags (see
-				   ISO15693_COMMAND_FLAGS_*) */
-  rfid_reader_t* reader;	/* RFID reader handler, the actual device. */
-  int error;			/* Errors set by iso15693 functions. */
-  char* device;			/* Path name of the device connecting
-				   to the RFID reader. */
-  int verbose;			/* Control verbosity level. */
+  rfid_reader_t* reader;	  /* RFID reader handler, the actual device. */
+  int error;			  /* Errors set by iso15693 functions. */
+  rfid_io_t* io;		  /* I/O physical device. */
+  int verbose;			  /* Control verbosity level. */
+  DB* system_information;         /* GET_SYSTEM_INFORMATION support. */
+  DBC* system_information_cursor; /* GET_SYSTEM_INFORMATION support. */
 } iso15693_t;
 
 /*
  * Abbreviations:
  *
- * AFI Application family identifier
+ * AFI Application Family Identifier
  *
  *     AFI represents the type of application targeted by the VCD and
  *     is used to extract from all the VICCs present only the VICCs
@@ -136,7 +140,7 @@ typedef struct iso15693 {
  *       return NO inventory
  *  }
  *
- * CRC Cyclic redundancy check
+ * CRC Cyclic Redundancy Check
  *
  *     The CRC shall be calculated in accordance with ISO/IEC
  *     13239. The initial register content shall be all ones:
@@ -153,7 +157,7 @@ typedef struct iso15693 {
  *     least significant byte first. Each byte is transmitted least
  *     significant bit first.
  *
- * DSFID Data storage format identifier
+ * DSFID Data Storage Format IDentifier
  *
  *       The Data storage format identifier indicates how the data is
  *       structured in the VICC memory. It may be programmed and
@@ -163,7 +167,7 @@ typedef struct iso15693 {
  *       by the VICC, the VICC shall respond with the value zero
  *       ('00').
  *
- * RFU Reserved for future use
+ * RFU Reserved for Future Use
  *
  * UID Unique identifier
  *
@@ -180,9 +184,9 @@ typedef struct iso15693 {
  *     |   E0   |  IC Mfg code | IC manufacturer serial number |
  *     +--------+--------------+-------------------------------+
  *
- * VCD Vicinity coupling device (or RFID reader)
+ * VCD Vicinity Coupling Device (or RFID reader)
  *
- * VICC Vicinity integrated circuit card (or RFID transponder or tag)
+ * VICC Vicinity Integrated Circuit Card (or RFID transponder or tag)
  */
 
 /*
@@ -201,6 +205,9 @@ typedef struct iso15693 {
 #define ISO15693_BLOCK_SECURITY_STATUS_SIZE	1
 #define ISO15693_VICC_MEMORY_SIZE_SIZE		2
 #define ISO15693_IC_REFERENCE_SIZE		1
+/* Not in ISO15693, added for INVENTORY command */
+#define NOT_ISO15693_VALID_DATA_FLAGS_SIZE	2
+#define NOT_ISO15693_COLLISION_FLAGS_SIZE	2
 
 /*
  * Header, trailer and overhead fixed sizes for command and response.
@@ -214,19 +221,25 @@ typedef struct iso15693 {
 #define ISO15693_RESPONSE_OVERHEAD		(ISO15693_RESPONSE_HEADER_SIZE + ISO15693_RESPONSE_TRAILER_SIZE)
 
 /*
- * Index of the fields that have a fixed position in the ISO-15693 frame.
+ * Index of the fields that have a fixed position in the ISO-15693 frames.
  */
 #define ISO15693_FLAGS_INDEX			0
 #define ISO15693_COMMAND_INDEX			1
 #define ISO15693_ERROR_INDEX			1
-#define ISO15693_DATA_INDEX			2
+#define ISO15693_COMMAND_DATA_INDEX		2
+#define ISO15693_RESPONSE_DATA_INDEX		1
+#define ISO15693_INFO_FLAGS_INDEX		1
+#define NOT_ISO15693_VALID_DATA_FLAGS_LSB_INDEX	1
+#define NOT_ISO15693_VALID_DATA_FLAGS_MSB_INDEX	2
+#define NOT_ISO15693_COLLISION_FLAGS_LSB_INDEX	3
+#define NOT_ISO15693_COLLISION_FLAGS_MSB_INDEX	4
 
 /*
  * VICC memory organization. The commands specified in ISO-15693-3
  * assume that the physical memory is organized in blocks (or pages)
  * of fixed size. Up to 256 blocks can be addressed. Block size
  * can be of up to 256 bits. This leads to a maximum memory capacity
- * of up to 8 kBytes (64 kBits).
+ * of up to 8 kBytes (64 kBits) (see ISO_IEC_15693-3 page 6).
  *
  * NOTE: The structure allows for future extension of the maximum
  * memory capacity. The commands described in this standard allow the
@@ -239,6 +252,13 @@ typedef struct iso15693 {
 #define ISO15693_BLOCK_COUNT_MAX	255
 #define ISO15693_BLOCK_SIZE_MAX		32   /* bytes */
 #define ISO15693_MEMORY_MAX		(ISO15693_BLOCK_COUNT_MAX * ISO15693_BLOCK_SIZE_MAX)
+/*
+ * Maximum size, in bytes, of a ISO-15693 command or answer.
+ * The extreme case of reference is a read_multiple_blocks with 
+ * a transponder that contains 255 blocks of 32 bytes.
+ * 
+ */
+#define ISO15693_BUFFER_MAX		(ISO15693_MEMORY_MAX * 2)
 
 /*
  * Bit description of the GET_SYSTEM_INFORMATION flags field
@@ -278,6 +298,7 @@ typedef struct {
  * AFI stands for Application family identifier.
  *  page 4, table 1 AFI coding
  */
+#define ISO15693_AFI_NONE			0x00
 #define ISO15693_AFI_TRANSPORT			0x10
 #define ISO15693_AFI_FINANCIAL			0x20
 #define ISO15693_AFI_IDENTIFICATION		0x30
@@ -378,16 +399,19 @@ typedef struct {
  * The second line shows the number of bytes.
  */
 
+#define ISO15693_ERROR_RESPONSE_SIZE	1
+#define ISO15693_ERROR_FRAME_SIZE	(ISO15693_FLAGS_SIZE + ISO15693_ERROR_RESPONSE_SIZE + ISO15693_CRC_SIZE)
+
 /*
  * Unset: No error.
  * Set: Error detected. Error code is in the "Error" field.
  */
-#define ISO15693_RESPONSE_FLAG_ERROR		0x01
+#define ISO15693_RESPONSE_FLAG_ERROR			0x01
 /*
  * Unset: No protocol format extension.
  * Set: Protocol format is extended. Reserved for future use.
  */
-#define ISO15693_RESPONSE_FLAG_EXTENSION	0x08
+#define ISO15693_RESPONSE_FLAG_EXTENSION		0x08
 
 /*
  * Response error code. when the ERROR flag is set by the VICC, the
@@ -397,54 +421,59 @@ typedef struct {
  * answer with the error code '0F' ("Error with no information
  * given").
  */
+#define ISO15693_SUCCESS				0x00
 /*
  * The command is not supported, i.e. the request code is not recognized.
  */
-#define ISO15693_ERROR_NOT_SUPPORTED		0x01
+#define ISO15693_ERROR_NOT_SUPPORTED			0x01
 /*
  * The command is not recognized, for example: a format error occured.
  */
-#define ISO15693_ERROR_NOT_RECOGNIZED		0x02
+#define ISO15693_ERROR_NOT_RECOGNIZED			0x02
 /*
  * The command option is not supported.
  */
-#define ISO15693_ERROR_OPTION_NOT_SUPPORTED	0x03
+#define ISO15693_ERROR_OPTION_NOT_SUPPORTED		0x03
 /*
  * Error with no information given or a specific error code is not supported.
  */
-#define ISO15693_ERROR_NO_INFORMATION		0x0F
+#define ISO15693_ERROR_NO_INFORMATION			0x0F
 /*
  * The specified block is not available (doesn't exist)
  */
-#define ISO15693_ERROR_BLOCK_NOT_AVAILABLE	0x10
+#define ISO15693_ERROR_BLOCK_NOT_AVAILABLE		0x10
 /*
  * The specified block is already locked and thus cannot be locked again.
  */
-#define ISO15693_ERROR_BLOCK_CANNOT_RELOCK	0x11
+#define ISO15693_ERROR_BLOCK_CANNOT_RELOCK		0x11
 /*
  * The specified block is locked and its content cannot be changed.
  */
-#define ISO15693_ERROR_BLOCK_CANNOT_WRITE	0x12
+#define ISO15693_ERROR_BLOCK_CANNOT_WRITE		0x12
 /*
  * The specified block was not successfully programmed.
  */
-#define ISO15693_ERROR_BLOCK_CANNOT_PROGRAM	0x13
+#define ISO15693_ERROR_BLOCK_CANNOT_PROGRAM		0x13
 /*
  * The specified block was not successfully locked.
  */
-#define ISO15693_ERROR_BLOCK_CANNOT_LOCK	0x14
+#define ISO15693_ERROR_BLOCK_CANNOT_LOCK		0x14
 
 /*
- * NOT part of ISO-15693. 
+ * Not part of ISO-15693. 
  */
 /*
  * The requested transponder (addressed mode) was not found.
  */
-#define ISO15693_ERROR_TRANSPONDER_NOT_FOUND	0x15
+#define NOT_ISO15693_ERROR_TRANSPONDER_NOT_FOUND	0x15
 /*
  * The flag combination is incorrect.
  */
-#define ISO15693_ERROR_INVALID_FLAGS		0x16
+#define NOT_ISO15693_ERROR_INVALID_FLAGS		0x16
+/*
+ * A collision occured after an INVENTORY request.
+ */
+#define NOT_ISO15693_ERROR_COLLISION			0x17
 
 /*
  * VICC states. A VICC can be in one of the 4 following states:
@@ -552,13 +581,54 @@ typedef struct {
  * --+--------------+----------------------------------------+--
  * ..|  8 bits      |        64 bits                         |..
  * --+--------------+----------------------------------------+--
+ *
+ * The drivers implementing access (s6350, m2xxh...) are required
+ * to implement the INVENTORY logic (as described by ISO/IEC 15693-3:2001(E)
+ * page 21,22,23) and and to return a cooked response :
+ * 
+ * Response:
+ * --+-------+---------+--------+----------------------------+--
+ * ..| Valid | Collide | DSFID  |         UID                |..
+ * --+-------+---------+--------+----------------------------+--
+ * ..|16 bits| 16 bits | 8 bits |      64 bits               |..
+ * --+-------+---------+--------+----------------------------+--
+ *                     \- repeat number of 1 bits in Valid --/
+ *
+ * Each bit set to one in the Valid field counts for 1 set of
+ * of DSFID/UID. If Valid == 0xFF, there are 16 DSFID/UID fields.
+ *
+ * Valid: This 16-bit field corresponds to whether valid data was
+ * received in the 16 possible Time Slots. Bits 0 to 7 of the LSB
+ * respectively correspond to Time Slots 1 to 8, while bits 0 to 7 of
+ * the MSB correspond to Time Slots 9 to 16 respectively. A set bit
+ * corresponds to valid data being received in that particular Time
+ * slot.
+ *
+ * Collide: This 16-bit field corresponding to whether a collision
+ * occurred in the 16 possible Time Slots. Bits 0 to 7 of the LSB
+ * respectively correspond to Time Slots 1 to 8, while bits 0 to 7 of
+ * the MSB correspond to Time Slots 9 to 16 respectively. A set bit
+ * corresponds to a collision being detected in that particular Time
+ * Slot.
+ *
  */
 #define ISO15693_COMMAND_INVENTORY				0x01
+#define ISO15693_COMMAND_INVENTORY_RESPONSE_SIZE		(ISO15693_DSFID_SIZE + ISO15693_UID_SIZE)
+
+/*
+ * Number of slots explored by the INVENTORY command.
+ */
+#define ISO15693_NB_SLOTS_DEFAULT		16
+#define ISO15693_NB_SLOTS_DEFAULT_BITS		4
+#define ISO15693_NB_SLOTS_ONE			1
+#define ISO15693_NB_SLOTS_ONE_BITS		1
 
 /*
  * Input parameter of the inventory command.
  */
 typedef struct {
+  u_int8_t one_slot;				 /* If set NB_SLOTS is 1 instead
+						    of 16 (default) */
   u_int8_t mask_length;				 /* Number of valid BITS in 
 						    mask_value. If 0, it is ignored. 
 						 */
@@ -575,12 +645,16 @@ typedef struct {
 } iso15693_inventory_t;
 
 /*
- * Output result of the inventory command (an array of these is actually
- * returned).
+ * Output result of the inventory command.
  */
 typedef struct {
-  u_int8_t dsfid;			/* Data Storage Format Identifier. */
-  u_int8_t uid[ISO15693_UID_SIZE];	/* Unique identifier. */
+  u_int16_t valid;			/* Validity of the data array. */
+  int nvalid;			        /* Total number of valid slots. */
+  u_int16_t collide;			/* Collision information. */
+  struct {
+    u_int8_t dsfid;			/* Data Storage Format Identifier. */
+    u_int8_t uid[ISO15693_UID_SIZE];	/* Unique identifier. */
+  } data[ISO15693_NB_SLOTS_DEFAULT];
 } iso15693_inventory_result_t;
 
 /*
@@ -598,7 +672,14 @@ typedef struct {
  * descriptions, matching the transponders currently detected by the
  * reader.
  */
-int iso15693_inventory(iso15693_t* iso15693, iso15693_inventory_t* inventory, iso15693_inventory_result_t** resultsp, int* results_lengthp);
+int iso15693_inventory(iso15693_t* iso15693, iso15693_inventory_t* inventory, iso15693_inventory_result_t* result);
+
+/*
+ * Full inventory of the RFID transponders currently in the field
+ * of the RFID reader. Contrary to the iso15693_inventory function,
+ * it resolves collisions.
+ */
+int iso15693_inventory_logic(struct rfid_reader* reader, iso15693_inventory_t inventory, rfid_transponders_t* transponders);
 
 /*
  * STAY_QUIET:
@@ -670,7 +751,7 @@ int iso15693_stay_quiet(iso15693_t* iso15693, u_int8_t* uid);
  * the caller to set <block->data> to a pointer with enough room to
  * store the content of a block.
  */
-int iso15693_read_single_block(iso15693_t* iso15693, rfid_transponder_t* transponder, rfid_block_t* block);
+int iso15693_read_single_block(iso15693_t* iso15693, const rfid_transponder_t* transponder, rfid_block_t* block);
 
 /*
  * WRITE_SINGLE_BLOCK:
@@ -1101,6 +1182,7 @@ int iso15693_lock_dsfid(iso15693_t* iso15693, const u_int8_t* uid);
  *   the IC manufacturer. 
  */
 #define ISO15693_COMMAND_GET_SYSTEM_INFORMATION			0x2B
+#define ISO15693_COMMAND_GET_SYSTEM_INFORMATION_RESPONSE_SIZE	(ISO15693_INFO_FLAGS_SIZE + ISO15693_UID_SIZE + ISO15693_DSFID_SIZE + ISO15693_AFI_SIZE + ISO15693_VICC_MEMORY_SIZE_SIZE + ISO15693_IC_REFERENCE_SIZE)
 
 /*
  * The GET_SYSTEM_INFORMATION is sent to the first available transponder
@@ -1142,19 +1224,28 @@ int iso15693_get_system_information(iso15693_t* iso15693, rfid_transponder_t* tr
  * --+-------------------------------------------------------+--
  *   \------------ repeat Number of blocks times ------------/
  */
-#define ISO15693_COMMAND_GET_MULTIPLE_BLOCK_SECURITY_STATUS	0x2C
-/* NOT IMPLEMENTED. */
-
+#define ISO15693_COMMAND_GET_MULTIPLE_BLOCK_SECURITY_STATUS		0x2C
 /*
- * Allocate a reader object that must be free'd by the
- * caller with reader->free_f(reader).
+ * Fill the security_status field of each <blocks> with information
+ * from the transponder, starting at <blocks>[0].block_number and
+ * successive blocks up to <blocks_length>.  The command is sent to
+ * the transponder with UID equal to <uid> (ISO15693_UID_SIZE
+ * bytes). If <uid> is all 0, the command is sent to whatever
+ * transponder is available.
  */
-int iso15693_reader_alloc(rfid_reader_t** reader);
+int iso15693_get_multiple_block_security_status(iso15693_t* iso15693, const u_int8_t* uid, rfid_block_t* blocks, int blocks_length);
 
 /*
  * Get access to iso15693 object.
  */
-#define reader_to_iso15693(reader) ((iso15693_t*)reader->private)
+#define iso15693_from_reader(reader) ((iso15693_t*)reader->private)
+
+/*
+ * Return the size, in bytes, of the response to a request. The
+ * transponder argument is only necessary when the response size depends
+ * on the block size of the transponder.
+ */
+int iso15693_response_size(rfid_reader_t* reader, char* request, int request_size, int* response_size, const rfid_transponder_t* transponder);
 
 /*
  * Set CRC16 at the end of a ISO-15693 frame (last two bytes).
@@ -1239,12 +1330,12 @@ int iso15693_crc16_check(u_int8_t* request, int request_length);
  * blocks part of the VICC of the transponder t.
  */
 #define iso15693_transponder_vicc_blocks_set(t,v) \
-        (iso15693_transponder_vicc_blocks_clear(t), (((iso15693_transponder_t*)(t)->private)->vicc_memory |= ((v) & 0xFF)))
+        (iso15693_transponder_vicc_blocks_clear(t), (((iso15693_transponder_t*)(t)->private)->vicc_memory |= ((v - 1) & 0xFF)))
 /*
  * Copy the blocks part of the VICC of the transponder to the v variable (u_int8_t).
  */
 #define iso15693_transponder_vicc_blocks_get(t,v) \
-	((v) = (((iso15693_transponder_t*)(t)->private)->vicc_memory & 0xFF))
+	((v) = (((iso15693_transponder_t*)(t)->private)->vicc_memory & 0xFF) + 1)
 
 /*
  * Set to 0 the bytes_per_block part of the VICC of the transponder t.
@@ -1256,13 +1347,13 @@ int iso15693_crc16_check(u_int8_t* request, int request_length);
  * bytes_per_block part of the VICC of the transponder t.
  */
 #define iso15693_transponder_vicc_bytes_per_block_set(t,v) \
-        (iso15693_transponder_vicc_bytes_per_block_clear(t), (((iso15693_transponder_t*)(t)->private)->vicc_memory |= (((v) & 0x1F) << 8)))
+        (iso15693_transponder_vicc_bytes_per_block_clear(t), (((iso15693_transponder_t*)(t)->private)->vicc_memory |= (((v - 1) & 0x1F) << 8)))
 /*
  * Copy the bytes_per_block part of the VICC of the transponder
  * to the v variable (u_int8_t).
  */
 #define iso15693_transponder_vicc_bytes_per_block_get(t,v) \
-	((v) = ((((iso15693_transponder_t*)(t)->private)->vicc_memory >> 8) & 0x1F))
+	((v) = ((((iso15693_transponder_t*)(t)->private)->vicc_memory >> 8) & 0x1F) + 1)
 
 /*
  * Copy the IC REFERENCE from the v memory area
