@@ -33,12 +33,14 @@
 #include <stdlib.h>
 #include <string.h>
 #include <time.h>
-#include <errno.h>
 
+#include <rfid_error.h>
 #include <s6350.h>
 #include <iso15693.h>
 #include <rfid.h>
 #include <rfid_qsort.h>
+#include <errno.h>
+#include <ltdl.h>
 
 static int block_signals(rfid_t* rfid)
 {
@@ -50,11 +52,13 @@ static int block_signals(rfid_t* rfid)
      sigaddset(&sigset, SIGQUIT) < 0 ||
      sigaddset(&sigset, SIGTERM) < 0) {
     rfid->error = errno;
+    RFID_ERROR_STACK(rfid->verbose);
     return -1;
   }
 
   if(sigprocmask(SIG_BLOCK, &sigset, &rfid->sigset) < 0) {
     rfid->error = errno;
+    RFID_ERROR_STACK(rfid->verbose);
     return -1;
   }
 
@@ -65,51 +69,236 @@ static int release_signals(rfid_t* rfid)
 {
   if(sigprocmask(SIG_SETMASK, &rfid->sigset, NULL) < 0) {
     rfid->error = errno;
+    RFID_ERROR_STACK(rfid->verbose);
     return -1;
   }
 
   if(sigemptyset(&rfid->sigset) < 0) {
     rfid->error = errno;
+    RFID_ERROR_STACK(rfid->verbose);
     return -1;
   }
 
   return 0;
 }
 
-int rfid_init(rfid_t* rfid)
+typedef struct {
+  const char** drivers;
+  int drivers_length;
+  const char** devices;
+  int devices_length;
+
+  rfid_reader_t** readers;
+  int readers_length;
+} rfid_probe_t;
+
+static int rfid_probe(rfid_t* rfid, rfid_probe_t* probe)
+{
+  rfid_reader_t* probers[probe->drivers_length];
+  int retval = -1;
+  int device_index;
+  int driver_index;
+  int found = 0;
+
+  probe->readers_length = 0;
+
+  memset(probers, '\0', sizeof(rfid_reader_t*) * probe->drivers_length);
+
+  for(driver_index = 0; driver_index < probe->drivers_length; driver_index++) {
+
+    /*
+     * We simply ignore drivers that cannot be loaded and we keep
+     * going with the others.
+     */
+    if(rfid_alloc(probe->drivers[driver_index], &probers[driver_index], rfid->verbose) < 0) {
+      
+      if(errno != RFID_ERROR_TRANSPONDER_NOT_FOUND) {
+	RFID_ERROR_STACK(rfid->verbose);
+	goto err;
+      }
+    } else {
+      if(rfid_io_alloc(&probers[driver_index]->io) < 0) {
+	rfid->error = errno;
+	RFID_ERROR_STACK(rfid->verbose);
+	goto err;
+      }
+
+      probers[driver_index]->verbose_f(probers[driver_index], rfid->verbose);
+
+      found++;
+    }
+  }
+
+  if(!found) {
+    rfid->error = RFID_ERROR_DRIVER_NOT_FOUND;
+    RFID_ERROR_STACK(rfid->verbose);
+    goto err;
+  }
+
+  for(device_index = 0; device_index < probe->devices_length; device_index++) {
+    int driver_index;
+    for(driver_index = 0; driver_index < probe->drivers_length; driver_index++) {
+      if(probers[driver_index]) {
+	rfid_reader_t* prober = probers[driver_index];
+
+	rfid_io_init(prober->io, probe->devices[device_index]);
+
+	if(prober->probe_f(prober, probe->drivers[driver_index]) < 0) {
+	  if(prober->error_f(prober) != RFID_ERROR_READER_NOT_FOUND) {
+	    rfid->error = prober->error_f(prober);
+	    if(rfid->verbose) fprintf(stderr, "encountered and error while attempting to probe device %s with driver %s\n", probe->devices[device_index], probe->drivers[driver_index]);
+	    RFID_ERROR_STACK(rfid->verbose);
+	    goto err;
+	  }
+	  if(rfid->verbose) {
+	    fprintf(stderr, "%s with %s: ", prober->io->device, prober->version_f(prober));
+	    fprintf(stderr, "nothing\n");
+	  }
+	} else {
+	  if(rfid->verbose) {
+	    fprintf(stderr, "%s with %s: ", prober->io->device, prober->version_f(prober));
+	    fprintf(stderr, "found RFID reader %s\n", prober->id_string);
+	  }
+	  if(prober->alloc_f(&probe->readers[probe->readers_length]) < 0) {
+	    rfid->error = prober->error_f(prober);
+	    RFID_ERROR_STACK(rfid->verbose);
+	    goto err;
+	  }
+	  strcpy(probe->readers[probe->readers_length]->drivers, prober->drivers);
+	  if(rfid_io_dup(prober->io, &probe->readers[probe->readers_length]->io) < 0) {
+	    rfid->error = errno;
+	    RFID_ERROR_STACK(rfid->verbose);
+	    goto err;
+	  }
+	  probe->readers_length++;
+	}
+      }
+    }
+  }
+
+  if(probe->readers_length > 0)
+    retval = 0;
+  else
+    rfid->error = RFID_ERROR_READER_NOT_FOUND;
+
+ err:
+
+  /*
+   * Release the probes.
+   */
+  for(driver_index = 0; driver_index < probe->drivers_length; driver_index++) {
+    if(probers[driver_index]) {
+      probers[driver_index]->end_f(probers[driver_index]);
+      rfid_io_free(probers[driver_index]->io);
+      rfid_free(probers[driver_index]);
+    }
+  }
+
+  if(retval < 0) {
+    /*
+     * Release the readers.
+     */
+    int i;
+    for(i = 0; i < probe->readers_length; i++) {
+      rfid_io_free(probe->readers[i]->io);
+      probe->readers[i]->free_f(probe->readers[i]);
+    }
+    probe->readers_length = 0;
+  }
+  
+  return retval;
+}
+
+int rfid_init(rfid_t* rfid, const char* driver, const char* device)
 {
   int retval = -1;
   rfid_reader_t* reader = 0;
+  const char* arg_drivers[1];
+  const char* arg_devices[1];
+  const char* default_devices[] = {
+    "/dev/ttyS0",
+    "/dev/ttyS1",
+    "/dev/ttyS2",
+    "/dev/ttyS3",
+    "/dev/ttyUSB0",
+    "/dev/ttyUSB1",
+    "/dev/ttyUSB2",
+    "/dev/ttyUSB3",
+  };
+  int devices_length = 8;
+  const char** devices = default_devices;
+
+  const char* default_drivers[] = {
+    "iso15693/s6350",
+    "iso15693/m2xxh",
+    "s6350",
+  };
+  int drivers_length = 3;
+  const char** drivers = default_drivers;
+
+  rfid_reader_t* readers[8 * 3] = { 0, };
+
+  rfid_probe_t probe = {
+    drivers,
+    drivers_length,
+    devices,
+    devices_length,
+
+    readers,
+    0
+  };
+  int i;
+
+  if(driver) {
+    arg_drivers[0] = driver;
+    probe.drivers = arg_drivers;
+    probe.drivers_length = 1;
+  }
+
+  if(device) {
+    arg_devices[0] = device;
+    probe.devices = arg_devices;
+    probe.devices_length = 1;
+  }
+  
 
   rfid->error = 0;
   if(sigemptyset(&rfid->sigset) < 0)
     return -1;
 
-  if(rfid->options & RFID_OPTION_ISO15693) {
-    if(iso15693_reader_alloc(&rfid->reader) < 0)
-      goto finish;
-  } else {
-    if(s6350_reader_alloc(&rfid->reader) < 0)
-      goto finish;
+  if(rfid_probe(rfid, &probe) < 0) {
+    RFID_ERROR_STACK(rfid->verbose);
+    goto err;
   }
 
-  reader = rfid->reader;
+  reader = rfid->reader = probe.readers[0];
 
-  if(rfid->options & RFID_OPTION_DUMP_FRAMES)
-    reader->verbose_f(reader, 2);
+  reader->verbose_f(reader, rfid->verbose);
 
-  reader->device = rfid->device;
+  for(i = 1; i < probe.readers_length; i++) {
+    probe.readers[i]->end_f(probe.readers[i]);
+    rfid_io_free(probe.readers[i]->io);
+    rfid_free(probe.readers[i]);
+  }
 
-  if(block_signals(rfid) < 0)
-    goto finish;
-  if(reader->init_f(reader) < 0)
-    goto finish;
+  rfid->options |= RFID_OPTION_FREE_IO;
+
+  if(block_signals(rfid) < 0) {
+    RFID_ERROR_STACK(rfid->verbose);
+    goto err;
+  }
+  if(reader->init_f(reader, reader->drivers) < 0) {
+    RFID_ERROR_STACK(rfid->verbose);
+    goto err;
+  }
 
   retval = 0;
 
- finish:
-  if(release_signals(rfid) < 0)
+ err:
+  if(release_signals(rfid) < 0) {
+    RFID_ERROR_STACK(rfid->verbose);
     return -1;
+  }
 
   return retval;
 }
@@ -155,7 +344,9 @@ int rfid_transponder_id_set(rfid_t* rfid, rfid_transponder_t* transponder, char*
 void rfid_end(rfid_t* rfid)
 {
   rfid->reader->end_f(rfid->reader);
-  rfid->reader->free_f(rfid->reader);
+  if(rfid->options & RFID_OPTION_FREE_IO)
+    rfid_io_free(rfid->reader->io);
+  rfid_free(rfid->reader);
 }
 
 char* rfid_strerror(rfid_t* rfid)
@@ -163,13 +354,13 @@ char* rfid_strerror(rfid_t* rfid)
   rfid_reader_t* reader = rfid->reader;
   char* message = (char*)malloc(512);
   
-  sprintf(message, "Device %s: ", reader->device);
-
   if(rfid->error) {
-      sprintf(message + strlen(message), "%d: %s", rfid->error, strerror(rfid->error));
+    sprintf(message, "%d: %s", rfid->error, rfid_internal_strerror(rfid->error));
+  } else if(!reader) {
+    sprintf(message, "unknow error");
   } else {
     char* error_string = reader->strerror_f(reader);
-    sprintf(message + strlen(message), "%s", error_string);
+    sprintf(message, "%s", error_string);
     free(error_string);
   }
 
@@ -187,39 +378,16 @@ int rfid_error(rfid_t* rfid)
   }
 }
 
-static void transponders_grow(rfid_t* rfid, rfid_transponder_t*** transpondersp, int* transponders_lengthp)
+static void transponders_free(rfid_t* rfid, rfid_transponders_t* transponders)
 {
-  rfid_reader_t* reader = rfid->reader;
-  int i;
-  rfid_transponder_t** transponders = *transpondersp;
-  int old_transponders_length = *transponders_lengthp;
-  int transponders_length = *transponders_lengthp ? ((*transponders_lengthp) * 2) : 8;
-  int size =  transponders_length * sizeof(rfid_transponder_t*);
-
-  if(old_transponders_length > 0) {
-      transponders = (rfid_transponder_t**)realloc(transponders, size);
-  } else {
-      transponders = (rfid_transponder_t**)malloc(size);
-  }
-
-  for(i = old_transponders_length; i < transponders_length; i++) {
-    reader->transponder_alloc_f(reader, transponders + i);
-  }
-
-  *transpondersp = transponders;
-  *transponders_lengthp = transponders_length;
-}
-
-static void transponders_free(rfid_t* rfid, rfid_transponder_t** transponders, int transponders_length)
-{
-  if(transponders) {
+  if(transponders->size > 0) {
     int i;
     rfid_reader_t* reader = rfid->reader;
 
-    for(i = 0; i < transponders_length; i++) {
-      reader->transponder_free_f(reader, transponders[i]);
-    }
-    free(transponders);
+    for(i = 0; i < transponders->size; i++)
+      reader->transponder_free_f(reader, transponders->list[i]);
+
+    free(transponders->list);
   }
 }
 
@@ -234,12 +402,6 @@ static void transponders_sort(rfid_t* rfid, rfid_transponder_t** transponders, i
   rfid_qsort(transponders, transponders_length, sizeof(rfid_transponder_t*), (rfid_qsort_cmp)transponders_cmp, rfid);
 }
 
-typedef struct {
-  int length;
-  int size;
-  rfid_transponder_t** transponders;
-} rfid_transponders_t;
-
 typedef struct {
   int events_mask;
   rfid_transponder_callback_t* transponder_callback;
@@ -277,7 +439,7 @@ static int transponders_diff(rfid_t* rfid, rfid_events_context_t* context, int*
     int compare;
 
     if(current_index < current->length && cache_index < cache->length) {
-      compare = reader->transponder_cmp_f(reader, current->transponders[current_index], cache->transponders[cache_index]);
+      compare = reader->transponder_cmp_f(reader, current->list[current_index], cache->list[cache_index]);
     } else if(current_index < current->length) {
       /*
        * What's left in current transponders is new because it matches
@@ -295,29 +457,30 @@ static int transponders_diff(rfid_t* rfid, rfid_events_context_t* context, int*
     if(compare < 0) {
       (*changedp)++;
       if(rfid->verbose == 1) fprintf(stderr, "N");
-      if(block_signals(rfid) < 0)
+      if(block_signals(rfid) < 0) {
 	return RFID_CALLBACK_ERROR;
+      }
       /*
        * Fill structural information about the transponder.
        */
-      if(reader->transponder_present_f(reader, current->transponders[current_index]) < 0) {
+      if(reader->transponder_present_f(reader, current->list[current_index]) < 0) {
 	/*
 	 * The state changed, try again.
 	 */
-	if(reader->error_f(reader) == ENOENT)
-	  rfid->error = EAGAIN;
+	if(reader->error_f(reader) == RFID_ERROR_TRANSPONDER_NOT_FOUND)
+	  rfid->error = RFID_ERROR_TRY_AGAIN;
 	release_signals(rfid);
 	return RFID_CALLBACK_ERROR;
       }
       /*
        * Read all data from it.
        */
-      if(reader->read_f(reader, 0, current->transponders[current_index]->blocks, current->transponders[current_index]) < 0) {
+      if(reader->read_f(reader, 0, current->list[current_index]->blocks, current->list[current_index]) < 0) {
 	/*
 	 * The state changed, try again.
 	 */
-	if(reader->error_f(reader) == ENOENT)
-	  rfid->error = EAGAIN;
+	if(reader->error_f(reader) == RFID_ERROR_TRANSPONDER_NOT_FOUND)
+	  rfid->error = RFID_ERROR_TRY_AGAIN;
 	release_signals(rfid);
 	return RFID_CALLBACK_ERROR;
       }
@@ -326,7 +489,7 @@ static int transponders_diff(rfid_t* rfid, rfid_events_context_t* context, int*
 	return RFID_CALLBACK_ERROR;
 
       if(events_mask & RFID_EVENT_NEW) {
-	if((retval = (*transponder_callback)(rfid, RFID_EVENT_NEW, current->transponders[current_index], data)) != 0) 
+	if((retval = (*transponder_callback)(rfid, RFID_EVENT_NEW, current->list[current_index], data)) != 0) 
 	  return retval;
       }
       current_index++;
@@ -334,7 +497,7 @@ static int transponders_diff(rfid_t* rfid, rfid_events_context_t* context, int*
       (*changedp)++;
       if(rfid->verbose == 1) fprintf(stderr, "D");
       if(events_mask & RFID_EVENT_DELETE) {
-	if((retval = (*transponder_callback)(rfid, RFID_EVENT_DELETE, cache->transponders[cache_index], data)) != 0) 
+	if((retval = (*transponder_callback)(rfid, RFID_EVENT_DELETE, cache->list[cache_index], data)) != 0) 
 	  return retval;
       }
       cache_index++;
@@ -343,10 +506,10 @@ static int transponders_diff(rfid_t* rfid, rfid_events_context_t* context, int*
        * Copy whatever is known about the transponder from the cache instead
        * of requesting data with a new present/read.
        */
-      reader->transponder_copy_f(reader, current->transponders[current_index], cache->transponders[cache_index]);
+      reader->transponder_copy_f(reader, current->list[current_index], cache->list[cache_index]);
       if(rfid->verbose == 1) fprintf(stderr, "A");
       if(events_mask & RFID_EVENT_AGAIN) {
-	if((retval = (*transponder_callback)(rfid, RFID_EVENT_AGAIN, current->transponders[current_index], data)) != 0) 
+	if((retval = (*transponder_callback)(rfid, RFID_EVENT_AGAIN, current->list[current_index], data)) != 0) 
 	  return retval;
       }
       current_index++;
@@ -377,12 +540,14 @@ static int event_loop(rfid_t* rfid, rfid_events_context_t* context)
      !transponder_callback) {
     if(rfid->verbose) fprintf(stderr, "rfid_event_loop: events_mask set to one of RFID_EVENT_NEW, RFID_EVENT_DELETE or RFID_EVENT_AGAIN but transponder_callback argument is null\n");
     rfid->error = EINVAL;
+    RFID_ERROR_STACK(rfid->verbose);
     return -1;
   }
 
   if((events_mask & RFID_EVENT_INVENTORY) && !inventory_callback) {
     if(rfid->verbose) fprintf(stderr, "rfid_event_loop: events_mask set to one of RFID_EVENT_INVENTORY but inventory_callback argument is null\n");
     rfid->error = EINVAL;
+    RFID_ERROR_STACK(rfid->verbose);
     return -1;
   }
   
@@ -391,43 +556,42 @@ static int event_loop(rfid_t* rfid, rfid_events_context_t* context)
   while(retval == RFID_CALLBACK_OK) {
     int changed = 0;
 
-    if(block_signals(rfid) < 0)
+    if(block_signals(rfid) < 0) {
+      RFID_ERROR_STACK(rfid->verbose);
       return -1;
+    }
 
-    if(reader->inventory_f(reader, &current->transponders, &current->length, &current->size) < 0) {
+    if(reader->inventory_f(reader, current) < 0) {
       int inventory_error = reader->error_f(reader);
-      if(release_signals(rfid) < 0)
+      if(release_signals(rfid) < 0) {
+	RFID_ERROR_STACK(rfid->verbose);
 	return -1;
-      /*
-       * Inventory tells us that it needs more room.
-       */
-      if(inventory_error == E2BIG) {
-	transponders_grow(rfid, &current->transponders, &current->size);
-	transponders_grow(rfid, &cache->transponders, &cache->size);
-	continue;
       }
       /*
        * Restart if a recoverable error occurs.
        */
-      if(inventory_error == EAGAIN) {
+      if(inventory_error == RFID_ERROR_TRY_AGAIN) {
 	if(rfid->verbose == 1) fprintf(stderr, "!");
 	continue;
       }
 
+      RFID_ERROR_STACK(rfid->verbose);
       return -1;
     }
 
-    if(release_signals(rfid) < 0)
+    if(release_signals(rfid) < 0) {
+      RFID_ERROR_STACK(rfid->verbose);
       return -1;
+    }
 
-    transponders_sort(rfid, current->transponders, current->length);
+    transponders_sort(rfid, current->list, current->length);
 
     if((retval = transponders_diff(rfid, context, &changed)) < 0) {
       /*
        * Recoverable error (typically the state changed in the middle of
        * the analysis of the current state).
        */
-      if(rfid_error(rfid) == EAGAIN) {
+      if(rfid_error(rfid) == RFID_ERROR_TRY_AGAIN) {
 	if(rfid->verbose == 1) fprintf(stderr, "!");
 	rfid->error = 0;
 	retval = RFID_CALLBACK_OK;
@@ -437,12 +601,15 @@ static int event_loop(rfid_t* rfid, rfid_events_context_t* context)
        * Everything else is a fatal error, either from the rfid
        * library or because a callback returned on error.
        */
+      RFID_ERROR_STACK(rfid->verbose);
       return -1;
     }
 
     if((retval == RFID_CALLBACK_OK) && changed && (events_mask & RFID_EVENT_INVENTORY)) {
-      if((retval = (*inventory_callback)(rfid, current->transponders, current->length, data)) < 0)
+      if((retval = (*inventory_callback)(rfid, current->list, current->length, data)) < 0) {
+	RFID_ERROR_STACK(rfid->verbose);
 	return -1;
+      }
     }
     
     if(rfid->verbose == 1) fprintf(stderr, "/");
@@ -482,16 +649,15 @@ int rfid_event_loop(rfid_t* rfid, int events_mask, rfid_transponder_callback_t*
   context.inventory_callback = inventory_callback;
   context.data = data;
 
-  transponders_grow(rfid, &context.current.transponders, &context.current.size);
-  transponders_grow(rfid, &context.cache.transponders, &context.cache.size);
-
   retval = event_loop(rfid, &context);
 
-  transponders_free(rfid, context.current.transponders, context.current.size);
-  transponders_free(rfid, context.cache.transponders, context.cache.size);
+  transponders_free(rfid, &context.current);
+  transponders_free(rfid, &context.cache);
 
-  if(release_signals(rfid) < 0)
+  if(release_signals(rfid) < 0) {
+    RFID_ERROR_STACK(rfid->verbose);
     return -1;
+  }
 
   return retval;
 }
@@ -508,24 +674,28 @@ int rfid_write(rfid_t* rfid, rfid_transponder_t* transponder)
     5000000,
   };
 
-  if(reader->transponder_alloc_f(reader, &current_transponder) < 0)
-    goto finish;
+  if(reader->transponder_alloc_f(reader, &current_transponder) < 0) {
+    RFID_ERROR_STACK(rfid->verbose);
+    goto err;
+  }
 
   rfid->error = 0;
 
   while(1) {
-    if(block_signals(rfid) < 0)
-      goto finish;
+    if(block_signals(rfid) < 0) {
+      RFID_ERROR_STACK(rfid->verbose);
+      goto err;
+    }
 
     if(reader->transponder_present_f(reader, current_transponder) < 0) {
 
-      if(reader->error_f(reader) == ENOENT) {
+      if(reader->error_f(reader) == RFID_ERROR_TRANSPONDER_NOT_FOUND) {
 	/*
 	 * No RFID transponder, try again.
 	 */
 	if(rfid->verbose == 1) fprintf(stderr, ".");
 
-      } else if(reader->error_f(reader) == EAGAIN) {
+      } else if(reader->error_f(reader) == RFID_ERROR_TRY_AGAIN) {
 	/*
 	 * Recoverable error, try again.
 	 */
@@ -535,7 +705,8 @@ int rfid_write(rfid_t* rfid, rfid_transponder_t* transponder)
 	/*
 	 * Unrecoverable error, abort.
 	 */
-	goto finish;
+	RFID_ERROR_STACK(rfid->verbose);
+	goto err;
       }
     } else {
       /*
@@ -544,43 +715,54 @@ int rfid_write(rfid_t* rfid, rfid_transponder_t* transponder)
       break;
     }
 
-    if(release_signals(rfid) < 0)
-      goto finish;
+    if(release_signals(rfid) < 0) {
+      RFID_ERROR_STACK(rfid->verbose);
+      goto err;
+    }
 
     nanosleep(&zzz, 0);
   }
 
   if(!reader->transponder_null_f(reader, transponder) &&
      reader->transponder_cmp_f(reader, current_transponder, transponder)) {
-    rfid->error = EXDEV;
-    goto finish;
+    rfid->error = RFID_ERROR_TRANSPONDER_NOT_FOUND;
+    RFID_ERROR_STACK(rfid->verbose);
+    goto err;
   }
 
   memcpy(current_transponder->data, transponder->data, transponder->data_length);
   current_transponder->data_length = transponder->data_length;
 
-  if(reader->write_f(reader, 0, current_transponder->blocks, current_transponder) < 0)
-    goto finish;
+  if(reader->write_f(reader, 0, current_transponder->blocks, current_transponder) < 0) {
+    RFID_ERROR_STACK(rfid->verbose);
+    goto err;
+  }
 
-  if(reader->transponder_copy_f(reader, transponder, current_transponder) < 0)
-    goto finish;
+  if(reader->transponder_copy_f(reader, transponder, current_transponder) < 0) {
+    RFID_ERROR_STACK(rfid->verbose);
+    goto err;
+  }
   
   retval = 0;
 
- finish:
+ err:
 
   if(current_transponder) {
-    if(reader->transponder_free_f(reader, current_transponder) < 0)
+    if(reader->transponder_free_f(reader, current_transponder) < 0) {
+      RFID_ERROR_STACK(rfid->verbose);
       return -1;
+    }
   }
 
-  if(release_signals(rfid) < 0)
+  if(release_signals(rfid) < 0) {
+    RFID_ERROR_STACK(rfid->verbose);
     return -1;
+  }
 
   return retval;
 }
 
 const char* rfid_version(void)
 {
-  return RFID_VERSION;
+  return "RFID "RFID_VERSION;
 }
