@@ -34,7 +34,6 @@
 #include <signal.h>
 #include <string.h>
 #include <memory.h>
-#include <errno.h>
 
 #include "rfid.h"
 
@@ -51,10 +50,10 @@ static RETSIGTYPE timeout_init(int sig)
 {
   fprintf(stderr, "\
 \n\
-FAILURE: Timeout attempting to establish communication with RFID reader at %s.\n\
+FAILURE: Timeout attempting to establish communication with RFID reader.\n\
 FAILURE: Make sure a RFID reader is plugged in and turned on.\n\
 \n\
-", rfid.device);
+");
   exit(-1);
 }
 
@@ -116,14 +115,16 @@ static int transponder_callback(rfid_t* rfid, int event, const rfid_transponder_
 int main(int argc, char** argv)
 {
   int timeout = 5;
+  int verbose = 0;
+  char* driver = RFID_DRIVER_GUESS;
+  char* device = RFID_DEVICE_GUESS;
 
   if(argc > 1)
-    rfid.device = argv[1];
-  else
-    rfid.device = "/dev/ttyS0";
-
+    driver = argv[1];
   if(argc > 2)
-    rfid_options(&rfid, RFID_OPTION_ISO15693);
+    device = argv[2];
+
+  rfid.verbose = verbose;
 
   if(signal(SIGALRM, timeout_init) < 0) {
     perror("signal(SIGALRM)");
@@ -131,12 +132,14 @@ int main(int argc, char** argv)
   }
 
   alarm(timeout);
-  if(rfid_init(&rfid) < 0) {
+  if(rfid_init(&rfid, driver, device) < 0) {
     report_error(&rfid, "rfid_init");
     return -1;
   }
   alarm(0);
 
+  rfid.reader->verbose_f(rfid.reader, rfid.verbose);
+
   if(signal(SIGALRM, timeout_write) < 0) {
     perror("signal(SIGALRM)");
     return -1;
@@ -174,8 +177,8 @@ Wrote string %s to transponder:\n\
     /* Fake manufacturer, unlikely to match an existing one to trigger failure. */
     rfid_transponder_id_set(&rfid, transponder, "1.2.3.4.5.6.7.8");
     if(rfid_write(&rfid, transponder) >= 0 ||
-       rfid.error != EXDEV) {
-      report_error(&rfid, "rfid_write (EXDEV was expected)");
+       rfid.error != RFID_ERROR_TRANSPONDER_NOT_FOUND) {
+      report_error(&rfid, "rfid_write (RFID_ERROR_TRANSPONDER_NOT_FOUND was expected)");
       rfid_transponder_free(&rfid, transponder);
       return -1;
     }
